# T-310/50 auf RaspberryPi PICO, Micropython
# SAS und Chiffrierdienst der DDR
# 07.07.2021 chiffrierdienst@kabelmail.de
# by Jörg Drobick
import rp2
from array import array
from random import randint
# T-310/50

ul_synchronfolge = 0 # Synchronfolge
s1_bit = array('B', [0,1,0,1,0,0,1,0,1,0,0,1,0,1,0,0,1,1,0,1,0,1,0,1,1,0,1,0,0,0,1,0,0,1,1,1,0,1,0,0,1,0,1,0,1,1,0,0,1,1,1,1,0,1,1,0,0,1,1,1,1,1,1,1,0,0,0,0,1,1,1,1,0,1,1,0,0,0,0,1,0,0,0,1,0,0,0,1,1,1,1,0,1,1,0,1,0,1,1,1,1,0,0,0,1,1,1,0,0,0,1,0,0,0,1,0,1,1,0,0])
s2_bit = array('B', [1,1,0,1,0,0,1,1,0,1,1,1,1,0,0,1,1,1,0,1,1,0,0,1,0,0,0,0,0,1,0,1,0,1,1,1,0,0,0,0,1,1,0,1,0,0,1,0,1,0,1,1,0,0,0,1,0,0,0,1,1,0,1,0,1,1,0,1,1,0,1,1,1,1,1,0,0,1,1,1,0,1,0,0,1,0,0,0,1,1,1,0,0,1,0,1,1,0,1,0,0,1,1,0,0,1,1,0,1,1,1,1,0,0,1,1,0,0,0,1])
# Standard U-Vektor
bo_U_Vektortemp = array ('B', [0, 0, 1, 1, 0, 1, 0, 0, 1, 1, 1, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 1, 1, 0, 1, 0, 0, 0, 1, 1])
ch_FsB = array('B', [0x4b, 0x65, 0x3d, 0x61, 0x20, 0x73, 0x69, 0x75, 0x3c, 0x64, 0x72, 0x6a, 0x6e, 0x66, 0x63, 0x6b, 0x74, 0x7a, 0x6c, 0x77, 0x68, 0x79, 0x70, 0x71, 0x6f, 0x62, 0x67, 0x23, 0x6d, 0x78, 0x76, 0x42]) # Teletype Latein
#ch_FsZ = array('B', [0x4b, 0x33, 0x3d, 0x2d, 0x20, 0x27, 0x38, 0x37, 0x3c, 0x23, 0x34, 0x6a, 0x2c, 0x46, 0x3a, 0x28, 0x35, 0x2b, 0x29, 0x32, 0x48, 0x36, 0x30, 0x31, 0x39, 0x3f, 0x47, 0x23, 0x2e, 0x2f, 0x3d, 0x42]) # Teletype Ziffern
ch_FsZ = array('H', [0x4b, 0x33, 0x3d, 0x2d,  0x20, 0x27, 0x38, 0x37, 0x3c,0xa4, 0x427, 0x44e,0x2c, 0x44d,0x3a, 0x28, 0x35, 0x2b, 0x29, 0x32, 0x429,0x36, 0x30, 0x31, 0x39, 0x3f, 0x448,0x23, 0x2e, 0x2f, 0x3d, 0x42])    # Teletyp Ziffern/Kyrillisch
ch_FsK = array('H', [0x4b, 0x435, 0x3d, 0x430, 0x20, 0x441,0x438,0x443,0x3c,0x434,0x440,0x439,0x43d,0x444,0x446,0x43a,0x442,0x437,0x43b,0x432,0x445,0x44b,0x43f,0x44f,0x43e,0x431,0x433,0x23, 0x43c,0x44c,0x436,0x42]) # Teletype Kyrillisch
# Maschienenbefehlsfolge
ch_MBF = array('B', [0x1f, 0x08, 0x02, 0x02])
# Zeiger auf S1/S2
b_zeiger = 0

def Schluesselpruefung():
    b_bloecke = 0   # 5 Blöcke a 25 bit
    while b_bloecke < 5:
        b_bits = 0
        bo_s1 = 0
        bo_s2 = 0
        while b_bits < 24:
            summe = (b_bloecke * 24) + b_bits
            bo_s1 ^= s1_bit[summe]
            bo_s2 ^= s2_bit[summe]
            b_bits += 1
        b_bloecke += 1
        if bo_s1 == 0: print("Fehler im Schlüsselabschnitt S1, Block:", b_bloecke)
        if bo_s2 == 0: print("Fehler im Schlüsselabschnitt S2, Block:", b_bloecke)
    if bo_s1 == 1: print("Schlüsselabschnitt S1 I.O.")
    if bo_s2 == 1: print("Schlüsselabschnitt S2 I.O.")

def Zufall():
    ul_kg = 0 # variable Kenngruppe
    while ul_kg == 0:
        ul_kg = randint(0, 0xffffff) # die ersten 24 bit
        ul_kg <<= 24
        ul_kg ^= randint(0, 0xffffff) # die zweitens 48 bit
        ul_kg <<= 24
        ul_kg ^= randint(0, 0xffff) # die drittens 64 bis
        ul_kg &= 0x1fffffffffffffff # Maskieren auf 61 Bit
    return ul_kg

def UVektorInit():
    global bo_U_Vektortemp
    global b_zeiger
    bo_U_Vektor = array ('B', [0, 0, 1, 1, 0, 1, 0, 0, 1, 1, 1, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 1, 1, 0, 1, 0, 0, 0, 1, 1])
    x = 0
    while ( x < 37):
        bo_U_Vektortemp[x] = bo_U_Vektor[x]
        x += 1
    b_zeiger = 0               # S1 und S0 wieder bei 0 beginnen

def KE():
    global b_zeiger
    global ul_synchronfolge
    global bo_U_Vektortemp
    b_fsBit = 0
    ui_wurmBit = 0
    b_LZS = array('B', [7, 4, 33, 30, 18, 36, 5, 35, 9, 16, 23, 26, 32, 12, 21, 1, 13, 25, 20, 8, 24, 15, 22, 29, 10, 28, 6, 0, 16, 4, 24, 12, 28, 32, 36, 20, 2]) # Zugelassen LZS-31, Langzeitschlüssel
    bo_Y_Vektor = array ('B', [0, 0, 0, 0, 0, 0, 0, 0, 0]) # Ergebnis und U-Vektoren, Beachtung zeitlicher Abläufe
    bo_Z = array ('B', [0, 0, 0, 0])     # Z0 ... Z3 (Z1 bis Z4)
    while b_fsBit < 13:                  # while 13 Runden # Achtung 13 Runden, Bit 0, 6, 12 und 13 sind ungenutzt
        b_runde = 0
        while b_runde < 127:             # for 0 to 127
            bo_tempS1 = s1_bit[b_zeiger] # S1 in T92 - Y1
            bo_tempS2 = s2_bit[b_zeiger] # S2 in Z0, T61 - Y13
            bo_Z[0] = Z(bo_tempS2, bo_U_Vektortemp[b_LZS[0]], bo_U_Vektortemp[b_LZS[1]], bo_U_Vektortemp[33], bo_U_Vektortemp[b_LZS[3]], bo_U_Vektortemp[b_LZS[4]])
            bo_Z[1] = Z(bo_U_Vektortemp[5], bo_U_Vektortemp[b_LZS[7]], bo_U_Vektortemp[9], bo_U_Vektortemp[b_LZS[9]], bo_U_Vektortemp[b_LZS[10]], bo_U_Vektortemp[b_LZS[11]])
            bo_Z[2] = Z(bo_U_Vektortemp[b_LZS[13]], bo_U_Vektortemp[21], bo_U_Vektortemp[b_LZS[15]], bo_U_Vektortemp[b_LZS[16]], bo_U_Vektortemp[25], bo_U_Vektortemp[b_LZS[18]])
            bo_Z[3] = Z(bo_U_Vektortemp[b_LZS[20]], bo_U_Vektortemp[b_LZS[21]], bo_U_Vektortemp[b_LZS[22]], bo_U_Vektortemp[29], bo_U_Vektortemp[b_LZS[24]], bo_U_Vektortemp[b_LZS[25]])
            bo_T = Rot_Fbit()
            bo_Y_Vektor[0] = bo_U_Vektortemp[b_LZS[35]] ^ bo_T  # Y33 = U20 xor f
            bo_T = bo_T ^ bo_Z[0]
            bo_Y_Vektor[1] = bo_U_Vektortemp[b_LZS[34]] ^ bo_T  # Y29 = U36 xor f xor Z0
            bo_T = bo_T ^ bo_U_Vektortemp[b_LZS[5]]             #       U36 entsprechend P''
            bo_Y_Vektor[2] = bo_U_Vektortemp[b_LZS[33]] ^ bo_T  # Y25 = U32 xor f xor Z0 xor U36
            bo_T = bo_T ^ bo_Z[1]
            bo_Y_Vektor[3] = bo_U_Vektortemp[b_LZS[32]] ^ bo_T  # Y21 = U28 xor f xor Z0 xor U36 xor Z1
            bo_T = bo_T ^ bo_U_Vektortemp[b_LZS[12]]            #       U12 entsprechend P''
            bo_Y_Vektor[4] = bo_U_Vektortemp[b_LZS[31]] ^ bo_T  # Y17 = U12 xor f xor Z0 xor U36 xor Z1 xor U32
            bo_T = bo_T ^ bo_tempS2 ^ bo_Z[2]
            bo_Y_Vektor[5] = bo_U_Vektortemp[b_LZS[30]] ^ bo_T  # Y13 = U24 xor f xor Z0 xor U36 xor Z1 xor U32 xor S2 xor Z2
            bo_T = bo_T ^ bo_U_Vektortemp[b_LZS[19]]            #       U8 entsprechend P''
            bo_Y_Vektor[6] = bo_U_Vektortemp[b_LZS[29]] ^ bo_T  # Y9  = U4  xor f xor Z0 xor U36 xor Z1 xor U32 xor S2 xor Z2 xor U8
            bo_T = bo_T ^ bo_Z[3]
            bo_Y_Vektor[7] = bo_U_Vektortemp[b_LZS[28]] ^ bo_T  # Y5  = U16 xor f xor Z0 xor U36 xor Z1 xor U32 xor S2 xor Z2 xor U8 xor Z3
            bo_T = bo_T ^ bo_U_Vektortemp[b_LZS[26]]            #       U6 entsprechend P''
            bo_Y_Vektor[8] = bo_tempS1 ^ bo_T                   # Y1  = S1  xor f xor Z0 xor U36 xor Z1 xor U32 xor S2 xor Z2 xor U8 xor Z3 xor U6
            bo_U_Vektortemp[36] = bo_U_Vektortemp[35] # U-Vektoren schieben
            bo_U_Vektortemp[35] = bo_U_Vektortemp[34]
            bo_U_Vektortemp[34] = bo_U_Vektortemp[33]
            bo_U_Vektortemp[33] = bo_Y_Vektor[0]  # Y33
            bo_U_Vektortemp[32] = bo_U_Vektortemp[31]
            bo_U_Vektortemp[31] = bo_U_Vektortemp[30]
            bo_U_Vektortemp[30] = bo_U_Vektortemp[29]
            bo_U_Vektortemp[29] = bo_Y_Vektor[1]  # Y29
            bo_U_Vektortemp[28] = bo_U_Vektortemp[27]
            bo_U_Vektortemp[27] = bo_U_Vektortemp[26]
            bo_U_Vektortemp[26] = bo_U_Vektortemp[25]
            bo_U_Vektortemp[25] = bo_Y_Vektor[2]  # Y25
            bo_U_Vektortemp[24] = bo_U_Vektortemp[23]
            bo_U_Vektortemp[23] = bo_U_Vektortemp[22]
            bo_U_Vektortemp[22] = bo_U_Vektortemp[21]
            bo_U_Vektortemp[21] = bo_Y_Vektor[3]  # Y21
            bo_U_Vektortemp[20] = bo_U_Vektortemp[19]
            bo_U_Vektortemp[19] = bo_U_Vektortemp[18]
            bo_U_Vektortemp[18] = bo_U_Vektortemp[17]
            bo_U_Vektortemp[17] = bo_Y_Vektor[4]  # Y17
            bo_U_Vektortemp[16] = bo_U_Vektortemp[15]
            bo_U_Vektortemp[15] = bo_U_Vektortemp[14]
            bo_U_Vektortemp[14] = bo_U_Vektortemp[13]
            bo_U_Vektortemp[13] = bo_Y_Vektor[5]  # Y13
            bo_U_Vektortemp[12] = bo_U_Vektortemp[11]
            bo_U_Vektortemp[11] = bo_U_Vektortemp[10]
            bo_U_Vektortemp[10] = bo_U_Vektortemp[9]
            bo_U_Vektortemp[9] = bo_Y_Vektor[6]   # Y9
            bo_U_Vektortemp[8] = bo_U_Vektortemp[7]
            bo_U_Vektortemp[7] = bo_U_Vektortemp[6]
            bo_U_Vektortemp[6] = bo_U_Vektortemp[5]
            bo_U_Vektortemp[5] = bo_Y_Vektor[7]   # Y5
            bo_U_Vektortemp[4] = bo_U_Vektortemp[3]
            bo_U_Vektortemp[3] = bo_U_Vektortemp[2]
            bo_U_Vektortemp[2] = bo_U_Vektortemp[1]
            bo_U_Vektortemp[1] = bo_Y_Vektor[8]  # Y1
            b_zeiger += 1                        # Zeiger auf nächstes Bit
            if b_zeiger > 119:
                b_zeiger = 0                     # Zeiger entspricht das rotieren des Schlüssels S1, S2
            b_runde += 1
        # for, 127 Runden!
        if bo_U_Vektortemp[b_LZS[36]] == 1:
            ui_wurmBit |= 1 << b_fsBit           # in der Reihenfolge 0...4, 5, 6...10
        b_fsBit += 1                             # nächstes bit
    # 13 Runden
    return ui_wurmBit                            # Rückgabe 11 bit Wurm

def Chiffrieren(b_klartext):
    ui_key = KE()
    # Schritt 1 ... 5, Schieberegister aufbauen
    b_SRV3 = 0x1f                                # Runden bis 0x1f in SRV2
    b_SRV2 = ui_key & 0x1f            # bit 0 ... bit 4 == 1-5 des Schlüssels, Schritt 1 ... 5, Schritt 6, Symetriewert bilden, rekursion
    while b_SRV2 < 0x1f & b_SRV2 > 0:
        b_SRV2 = Rekursion(b_SRV2)
        b_SRV3 = Rekursion(b_SRV3)
    b_SRV2 = b_SRV3                                  # b_SRV3 nach b_SRV2 kopieren
    b_SRV3 = b_klartext ^ ((ui_key & 0x07c0) >> 6)   # Additionsreihe bit 6 bis 1,0 (7 ... 1,1) XOR Ktxt (5bit), Schritt 7 ... 1,1
    while b_SRV2 < 0x1f & b_SRV2 > 0:                # bit 0 bis 4 (1-5)
        b_SRV2 = Rekursion(b_SRV2)
        b_SRV3 = Rekursion(b_SRV3)
    return b_SRV3   # Ausgabe Chiffriertext

def Dechiffrieren(b_crypt):
    ui_key = KE()
    b_SRV3 = b_crypt
    b_SRV2 = ui_key & 0x1f           # bit 0 bis 4 (1-5)
    while (b_SRV2 < 0x1f) & (b_SRV2 > 0):
    # Rekursion der Additionsreihe in SRV2 und des GTX in SRV3
        b_SRV2 = Rekursion(b_SRV2)
        b_SRV3 = Rekursion(b_SRV3)   # Z gebildet
    b_SRV3 ^= (ui_key >> 6) & 0x1f   # XOR Additionsreihe bit 6 bis 11 (7 ... 11) mit dem Zwischentext Z
    return b_SRV3

def Z(b_1, b_2, b_3, b_4, b_5, b_6):
    bo_temp = 1 ^ b_1 ^ b_5 ^ b_6 ^ (b_1 & b_4)
    bo_temp ^= (b_2 & b_3) ^ (b_2 & b_5) ^ (b_4 & b_5) ^ (b_5 & b_6)
    bo_temp ^= (b_1 & b_3 & b_4) ^ (b_1 & b_3 & b_6) ^ (b_1 & b_4 & b_5)
    bo_temp ^= (b_2 & b_3 & b_6) ^ (b_2 & b_4 & b_6) ^ (b_3 & b_5 & b_6)
    bo_temp ^= (b_1 & b_2 & b_3 & b_4) ^ (b_1 & b_2 & b_3 & b_5)
    bo_temp ^= (b_1 & b_2 & b_5 & b_6) ^ (b_2 & b_3 & b_4 & b_6)
    bo_temp ^= (b_1 & b_2 & b_3 & b_4 & b_5) ^ (b_1 & b_3 & b_4 & b_5 & b_6)
    return (bo_temp)

def Rekursion(b_Register):
    b_Bit0 = 0x01
    b_Bit3 = 0x04
    b_5Bitmaske = 0x1f            # Bitmaske aller werte innerhalb von 0 ... 31
    b_ret = b_Register            # übergabe der Eingabe an das Ausgabebyte
    bo_xor = ((b_Bit0 & b_ret) != 0) ^ ((b_Bit3 & b_ret) != 0)
    b_ret >>= 1                   # bo_XOR bildet rekusrionsbit aus Bit 5 und Bit 3
    if bo_xor == True:
        b_ret |= 0x10
    b_ret &= b_5Bitmaske
    return b_ret

def Synchronfolge(ul_Para, b_Runden):
    b_i = 0
    while b_Runden > b_i :
    # Umformung durch Bit 0 xor 1 xor 2 xor 5
        ul_Para = Rot_Ffolge(ul_Para)
        b_i += 1
    return ul_Para

# rotiere f-Folge, Übergabe- und Rückgabe Synchronfolge
def Rot_Ffolge(ul_Paramter):
    bo_m2 = ((ul_Paramter & 0x1) != 0) ^ ((ul_Paramter & 0x2) != 0) ^ ((ul_Paramter & 0x4) != 0) ^ ((ul_Paramter & 0x20) != 0) # Umformung durch Bit 0 xor 1 xor 2 xor 5
    ul_Paramter >>= 1     # Shift 1, von 60 nach 0
    if (bo_m2 == 1):
        ul_Paramter |= 0x1000000000000000   # bit 61
    return ul_Paramter

# rotiere f-Folge, Rückgabe Übertrag M2
def Rot_Fbit():
    global ul_synchronfolge
    bo_m2 = ((ul_synchronfolge & 0x1) != 0) ^ ((ul_synchronfolge & 0x2) != 0) ^ ((ul_synchronfolge & 0x4) != 0) ^ ((ul_synchronfolge & 0x20) != 0) # Umformung durch Bit 0 xor 1 xor 2 xor 5
    ul_synchronfolge >>= 1     # Shift 1, von 60 nach 0
    if (bo_m2 == 1):
        ul_synchronfolge |= 0x1000000000000000   # bit 61
    return bo_m2

# Main Routine
synch = ""
klartext = ""
ul_temp = Zufall()
ul_lese = ul_temp              # in Zufall bereits auf 61 bit maskiert
temp = 0
check = 0
if check == 0:                 # debug -> Senden
    print("Chiffrieren MBF")
    while (temp < 12):         # 60 bit,  12 * 5bit + 1bit; bit 0 ... 59
        b_lese = ul_lese & 0x1f    # nur das erste, 5 bit lesen
        synch += (chr(ch_FsB[b_lese]))    # speichern
        ul_lese >>= 5          # nächsten 5 bit holen
        temp += 1
    ul_temp = Synchronfolge(ul_temp, 60)  # weiter rotieren für Fs Zeichen 13 ... 24
    ul_lese = ul_temp               # kopieren
    temp = 0
    while (temp < 12):              # 120 bit, 12 * 5bit + 1bit; bit 0 ... 59
        b_lese = ul_lese & 0x1f     # nur das erste, 5 bit lesen
        synch += (chr(ch_FsB[b_lese]))    # speichern
        ul_lese >>= 5               # nächsten 5 bit holen
        temp += 1
    ul_temp = Synchronfolge(ul_temp, 4)   # weiter rotieren für Fs Zeichen 13 ... 20
    ul_synchronfolge |= ul_temp
    b_lese = (ul_temp >> 56) & 0x1f # nur das erste, 5 bit lesen
    synch += (chr(ch_FsB[b_lese]))  # speichern
    temp = 0
    Schluesselpruefung()
    print("bbbb" + synch + "kkkk")
    UVektorInit()                   # U-Vektor auf Anfang und S1 und S2 wieder bei 0 beginnen
    for i in ch_MBF:
        print (chr(ch_FsB[(Chiffrieren(i))]))
    check = 1                       # debug -> Empfangen
if check == 1:
    print("Dechiffrieren Testvektor")
    gtx = "bbbbqc<r=hzil<vpk <BksqnceKBBkkkkBqtvpnne<n zeflrwrblssredpusxgvzsnm gbi welg"

    if ((gtx[0:4] == "bbbb") & (gtx[29:33] == "kkkk")):
        # Synchronfolge aus Text extrahieren
        syF = gtx[4:29]
        # in Binär umwandeln
        ul_synchronfolge = 0
        z = 0
        while (z < 25):             # 25 Fs-Zeichen 5 bit
            for i in range(0, 32):  # suche in Array 32 Zeichen
                if (chr(ch_FsB[i]) == syF[24 - z]):
                    ul_synchronfolge |= i
                    if (z < 24):
                        ul_synchronfolge <<= 5
            z += 1                  # nächstes Element, erstmal 128bit, statt 61 bit
        ul_Testfolge = (ul_synchronfolge & 0x1fffffffffffffff)
        ul_synchronfolge >>= 64
        if (ul_synchronfolge == Synchronfolge(ul_Testfolge, 64)):  # Testergebnis: 0x1ff802e65caffa08
            print("Synchronfolge I.O. " + hex(ul_synchronfolge))
        else:
            print("Fehler in der Synchronfolge")
    #    Dechiffrieren(Text)
    else:
        print("Start- Endfolge der Synchronfolge falsch")
    Schluesselpruefung()
    # Dechiffrieren Testvektor
    print("Dechiffrierter Text: \nbbbbkkkk")
    z = 33                     # nach BBBB...KKKK
    UVektorInit()              # U-Vektor auf Anfang, S1 und S2 wieder bei 0 beginnen
    register = 1
    while (z < 76):
        for i in range(0, 32):
            if (chr(ch_FsB[i]) == gtx[z]):
                erg = Dechiffrieren(i)
                if erg == 0x1f:
                    register = 1
                elif erg == 0x1b:
                    register = 2
                elif erg == 0x00:
                    register = 3
                # Drucken in entsprechendem Zeichensatz, Latein/Ziffern/Kyrillisch
                if register == 1:
                    print (chr(ch_FsB[erg]))
                elif register == 2:
                    print (chr(ch_FsZ[erg]))
                elif register == 3:
                    print (chr(ch_FsK[erg]))
        z += 1
